@@ -20,22 +20,32 @@
 Response of ListBuckets, ListObjects, ListObjectsV2 and ListObjectVersions API.
 """
 
-from __future__ import absolute_import
+from __future__ import absolute_import, annotations
 
 import base64
-import datetime
 import json
 from collections import OrderedDict
+from datetime import datetime
 from enum import Enum
+from typing import Any, List, Tuple, Type, TypeVar, cast
 from urllib.parse import unquote_plus
 from xml.etree import ElementTree as ET
 
+from urllib3._collections import HTTPHeaderDict
+
+try:
+    from urllib3.response import BaseHTTPResponse  # type: ignore[attr-defined]
+except ImportError:
+    from urllib3.response import HTTPResponse as BaseHTTPResponse
+
+from .commonconfig import Tags
 from .credentials import Credentials
 from .helpers import check_bucket_name
 from .signer import get_credential_string, post_presign_v4
 from .time import from_iso8601utc, to_amz_date, to_iso8601utc
 from .xml import find, findall, findtext
 
+JSONDecodeError: type[ValueError]
 try:
     from json.decoder import JSONDecodeError
 except ImportError:
@@ -45,17 +55,17 @@ except ImportError:
 class Bucket:
     """Bucket information."""
 
-    def __init__(self, name, creation_date):
+    def __init__(self, name: str, creation_date: datetime | None):
         self._name = name
         self._creation_date = creation_date
 
     @property
-    def name(self):
+    def name(self) -> str:
         """Get name."""
         return self._name
 
     @property
-    def creation_date(self):
+    def creation_date(self) -> datetime | None:
         """Get creation date."""
         return self._creation_date
 
@@ -76,10 +86,13 @@ class Bucket:
         return hash(self.name)
 
 
+A = TypeVar("A", bound="ListAllMyBucketsResult")
+
+
 class ListAllMyBucketsResult:
     """LissBuckets API result."""
 
-    def __init__(self, buckets):
+    def __init__(self, buckets: list[Bucket]):
         self._buckets = buckets
 
     @property
@@ -88,32 +101,44 @@ class ListAllMyBucketsResult:
         return self._buckets
 
     @classmethod
-    def fromxml(cls, element):
+    def fromxml(cls: Type[A], element: ET.Element) -> A:
         """Create new object with values from XML element."""
-        element = find(element, "Buckets")
+        element = cast(ET.Element, find(element, "Buckets", True))
         buckets = []
-        if element is not None:
-            elements = findall(element, "Bucket")
-            for bucket in elements:
-                name = findtext(bucket, "Name", True)
-                creation_date = findtext(bucket, "CreationDate")
-                if creation_date:
-                    creation_date = from_iso8601utc(creation_date)
-                buckets.append(Bucket(name, creation_date))
+        elements = findall(element, "Bucket")
+        for bucket in elements:
+            name = cast(str, findtext(bucket, "Name", True))
+            creation_date = findtext(bucket, "CreationDate")
+            buckets.append(Bucket(
+                name,
+                from_iso8601utc(creation_date) if creation_date else None,
+            ))
         return cls(buckets)
 
 
+B = TypeVar("B", bound="Object")
+
+
 class Object:
     """Object information."""
 
-    def __init__(self,  # pylint: disable=too-many-arguments
-                 bucket_name,
-                 object_name,
-                 last_modified=None, etag=None,
-                 size=None, metadata=None,
-                 version_id=None, is_latest=None, storage_class=None,
-                 owner_id=None, owner_name=None, content_type=None,
-                 is_delete_marker=False):
+    def __init__(  # pylint: disable=too-many-arguments
+            self,
+            bucket_name: str,
+            object_name: str | None,
+            last_modified: datetime | None = None,
+            etag: str | None = None,
+            size: int | None = None,
+            metadata: dict[str, str] | HTTPHeaderDict | None = None,
+            version_id: str | None = None,
+            is_latest: str | None = None,
+            storage_class: str | None = None,
+            owner_id: str | None = None,
+            owner_name: str | None = None,
+            content_type: str | None = None,
+            is_delete_marker: bool = False,
+            tags: Tags | None = None,
+    ):
         self._bucket_name = bucket_name
         self._object_name = object_name
         self._last_modified = last_modified
@@ -127,80 +152,93 @@ class Object:
         self._owner_name = owner_name
         self._content_type = content_type
         self._is_delete_marker = is_delete_marker
+        self._tags = tags
 
     @property
-    def bucket_name(self):
+    def bucket_name(self) -> str:
         """Get bucket name."""
         return self._bucket_name
 
     @property
-    def object_name(self):
+    def object_name(self) -> str | None:
         """Get object name."""
         return self._object_name
 
     @property
-    def is_dir(self):
+    def is_dir(self) -> bool:
         """Get whether this key is a directory."""
-        return self._object_name.endswith("/")
+        return (
+            self._object_name is not None and self._object_name.endswith("/")
+        )
 
     @property
-    def last_modified(self):
+    def last_modified(self) -> datetime | None:
         """Get last modified time."""
         return self._last_modified
 
     @property
-    def etag(self):
+    def etag(self) -> str | None:
         """Get etag."""
         return self._etag
 
     @property
-    def size(self):
+    def size(self) -> int | None:
         """Get size."""
         return self._size
 
     @property
-    def metadata(self):
+    def metadata(self) -> dict[str, str] | HTTPHeaderDict | None:
         """Get metadata."""
         return self._metadata
 
     @property
-    def version_id(self):
+    def version_id(self) -> str | None:
         """Get version ID."""
         return self._version_id
 
     @property
-    def is_latest(self):
+    def is_latest(self) -> str | None:
         """Get is-latest flag."""
         return self._is_latest
 
     @property
-    def storage_class(self):
+    def storage_class(self) -> str | None:
         """Get storage class."""
         return self._storage_class
 
     @property
-    def owner_id(self):
+    def owner_id(self) -> str | None:
         """Get owner ID."""
         return self._owner_id
 
     @property
-    def owner_name(self):
+    def owner_name(self) -> str | None:
         """Get owner name."""
         return self._owner_name
 
     @property
-    def is_delete_marker(self):
+    def is_delete_marker(self) -> bool:
         """Get whether this key is a delete marker."""
         return self._is_delete_marker
 
     @property
-    def content_type(self):
+    def content_type(self) -> str | None:
         """Get content type."""
         return self._content_type
 
+    @property
+    def tags(self) -> Tags | None:
+        """Get the tags"""
+        return self._tags
+
     @classmethod
-    def fromxml(cls, element, bucket_name, is_delete_marker=False,
-                encoding_type=None):
+    def fromxml(
+            cls: Type[B],
+            element: ET.Element,
+            bucket_name: str,
+            is_delete_marker: bool = False,
+            encoding_type: str | None = None,
+    ) -> B:
         """Create new object with values from XML element."""
         tag = findtext(element, "LastModified")
         last_modified = None if tag is None else from_iso8601utc(tag)
@@ -211,22 +249,33 @@ class Object:
         tag = findtext(element, "Size")
         size = None if tag is None else int(tag)
 
-        tag = find(element, "Owner")
+        elem = find(element, "Owner")
         owner_id, owner_name = (
-            (None, None) if tag is None
-            else (findtext(tag, "ID"), findtext(tag, "DisplayName"))
+            (None, None) if elem is None
+            else (findtext(elem, "ID"), findtext(elem, "DisplayName"))
         )
 
-        tag = find(element, "UserMetadata") or []
-        metadata = {}
-        for child in tag:
+        elems: ET.Element | list = find(element, "UserMetadata") or []
+        metadata: dict[str, str] = {}
+        for child in elems:
             key = child.tag.split("}")[1] if "}" in child.tag else child.tag
-            metadata[key] = child.text
+            metadata[key] = child.text or ""
 
-        object_name = findtext(element, "Key", True)
+        object_name = cast(str, findtext(element, "Key", True))
         if encoding_type == "url":
             object_name = unquote_plus(object_name)
 
+        tags_text = findtext(element, "UserTags")
+        tags: Tags | None = None
+        if tags_text:
+            tags = Tags.new_object_tags()
+            tags.update(
+                cast(
+                    List[Tuple[Any, Any]],
+                    [tokens.split("=") for tokens in tags_text.split("&")],
+                ),
+            )
+
         return cls(
             bucket_name,
             object_name,
@@ -240,13 +289,17 @@ class Object:
             owner_name=owner_name,
             metadata=metadata,
             is_delete_marker=is_delete_marker,
+            tags=tags
         )
 
 
-def parse_list_objects(response, bucket_name=None):
+def parse_list_objects(
+        response: BaseHTTPResponse,
+        bucket_name: str | None = None,
+) -> tuple[list[Object], bool, str | None, str | None]:
     """Parse ListObjects/ListObjectsV2/ListObjectVersions response."""
     element = ET.fromstring(response.data.decode())
-    bucket_name = findtext(element, "Name", True)
+    bucket_name = cast(str, findtext(element, "Name", True))
     encoding_type = findtext(element, "EncodingType")
     elements = findall(element, "Contents")
     objects = [
@@ -264,7 +317,7 @@ def parse_list_objects(response, bucket_name=None):
     elements = findall(element, "CommonPrefixes")
     objects += [
         Object(
-            bucket_name, unquote_plus(findtext(tag, "Prefix", True))
+            bucket_name, unquote_plus(findtext(tag, "Prefix", True) or "")
             if encoding_type == "url" else findtext(tag, "Prefix", True)
         ) for tag in elements
     ]
@@ -296,7 +349,7 @@ def parse_list_objects(response, bucket_name=None):
 class CompleteMultipartUploadResult:
     """CompleteMultipartUpload API result."""
 
-    def __init__(self, response):
+    def __init__(self, response: BaseHTTPResponse):
         element = ET.fromstring(response.data.decode())
         self._bucket_name = findtext(element, "Bucket")
         self._object_name = findtext(element, "Key")
@@ -308,83 +361,92 @@ class CompleteMultipartUploadResult:
         self._http_headers = response.headers
 
     @property
-    def bucket_name(self):
+    def bucket_name(self) -> str | None:
         """Get bucket name."""
         return self._bucket_name
 
     @property
-    def object_name(self):
+    def object_name(self) -> str | None:
         """Get object name."""
         return self._object_name
 
     @property
-    def location(self):
+    def location(self) -> str | None:
         """Get location."""
         return self._location
 
     @property
-    def etag(self):
+    def etag(self) -> str | None:
         """Get etag."""
         return self._etag
 
     @property
-    def version_id(self):
+    def version_id(self) -> str | None:
         """Get version ID."""
         return self._version_id
 
     @property
-    def http_headers(self):
+    def http_headers(self) -> HTTPHeaderDict:
         """Get HTTP headers."""
         return self._http_headers
 
 
+C = TypeVar("C", bound="Part")
+
+
 class Part:
     """Part information of a multipart upload."""
 
-    def __init__(self, part_number, etag, last_modified=None, size=None):
+    def __init__(
+            self,
+            part_number: int,
+            etag: str,
+            last_modified: datetime | None = None,
+            size: int | None = None,
+    ):
         self._part_number = part_number
         self._etag = etag
         self._last_modified = last_modified
         self._size = size
 
     @property
-    def part_number(self):
+    def part_number(self) -> int:
         """Get part number. """
         return self._part_number
 
     @property
-    def etag(self):
+    def etag(self) -> str:
         """Get etag."""
         return self._etag
 
     @property
-    def last_modified(self):
+    def last_modified(self) -> datetime | None:
         """Get last-modified."""
         return self._last_modified
 
     @property
-    def size(self):
+    def size(self) -> int | None:
         """Get size."""
         return self._size
 
     @classmethod
-    def fromxml(cls, element):
+    def fromxml(cls: Type[C], element: ET.Element) -> C:
         """Create new object with values from XML element."""
-        part_number = findtext(element, "PartNumber", True)
-        etag = findtext(element, "ETag", True)
+        part_number = int(cast(str, findtext(element, "PartNumber", True)))
+        etag = cast(str, findtext(element, "ETag", True))
         etag = etag.replace('"', "")
         tag = findtext(element, "LastModified")
         last_modified = None if tag is None else from_iso8601utc(tag)
         size = findtext(element, "Size")
-        if size:
-            size = int(size)
-        return cls(part_number, etag, last_modified, size)
+        return cls(
+            part_number, etag, last_modified, int(size) if size else None,
+        )
 
 
 class ListPartsResult:
     """ListParts API result."""
 
-    def __init__(self, response):
+    def __init__(self, response: BaseHTTPResponse):
         element = ET.fromstring(response.data.decode())
         self._bucket_name = findtext(element, "Bucket")
         self._object_name = findtext(element, "Key")
@@ -404,78 +466,75 @@ class ListPartsResult:
         )
         self._storage_class = findtext(element, "StorageClass")
         self._part_number_marker = findtext(element, "PartNumberMarker")
-        self._next_part_number_marker = findtext(
-            element, "NextPartNumberMarker",
+        next_part_number_marker = findtext(element, "NextPartNumberMarker")
+        self._next_part_number_marker = (
+            int(next_part_number_marker) if next_part_number_marker else None
         )
-        if self._next_part_number_marker:
-            self._next_part_number_marker = int(self._next_part_number_marker)
-        self._max_parts = findtext(element, "MaxParts")
-        if self._max_parts:
-            self._max_parts = int(self._max_parts)
-        self._is_truncated = findtext(element, "IsTruncated")
+        max_parts = findtext(element, "MaxParts")
+        self._max_parts = int(max_parts) if max_parts else None
+        is_truncated = findtext(element, "IsTruncated")
         self._is_truncated = (
-            self._is_truncated is not None and
-            self._is_truncated.lower() == "true"
+            is_truncated is not None and is_truncated.lower() == "true"
         )
         self._parts = [Part.fromxml(tag) for tag in findall(element, "Part")]
 
     @property
-    def bucket_name(self):
+    def bucket_name(self) -> str | None:
         """Get bucket name."""
         return self._bucket_name
 
     @property
-    def object_name(self):
+    def object_name(self) -> str | None:
         """Get object name."""
         return self._object_name
 
     @property
-    def initiator_id(self):
+    def initiator_id(self) -> str | None:
         """Get initiator ID."""
         return self._initiator_id
 
     @property
-    def initator_name(self):
+    def initator_name(self) -> str | None:
         """Get initiator name."""
         return self._initiator_name
 
     @property
-    def owner_id(self):
+    def owner_id(self) -> str | None:
         """Get owner ID."""
         return self._owner_id
 
     @property
-    def owner_name(self):
+    def owner_name(self) -> str | None:
         """Get owner name."""
         return self._owner_name
 
     @property
-    def storage_class(self):
+    def storage_class(self) -> str | None:
         """Get storage class."""
         return self._storage_class
 
     @property
-    def part_number_marker(self):
+    def part_number_marker(self) -> str | None:
         """Get part number marker."""
         return self._part_number_marker
 
     @property
-    def next_part_number_marker(self):
+    def next_part_number_marker(self) -> int | None:
         """Get next part number marker."""
         return self._next_part_number_marker
 
     @property
-    def max_parts(self):
+    def max_parts(self) -> int | None:
         """Get max parts."""
         return self._max_parts
 
     @property
-    def is_truncated(self):
+    def is_truncated(self) -> bool:
         """Get is-truncated flag."""
         return self._is_truncated
 
     @property
-    def parts(self):
+    def parts(self) -> list[Part]:
         """Get parts."""
         return self._parts
 
@@ -483,12 +542,12 @@ class ListPartsResult:
 class Upload:
     """ Upload information of a multipart upload."""
 
-    def __init__(self, element, encoding_type=None):
+    def __init__(self, element: ET.Element, encoding_type: str | None = None):
         self._encoding_type = encoding_type
-        self._object_name = findtext(element, "Key", True)
+        object_name = cast(str, findtext(element, "Key", True))
         self._object_name = (
-            unquote_plus(self._object_name) if self._encoding_type == "url"
-            else self._object_name
+            unquote_plus(object_name) if self._encoding_type == "url"
+            else object_name
         )
         self._upload_id = findtext(element, "UploadId")
         tag = find(element, "Initiator")
@@ -506,47 +565,48 @@ class Upload:
             None if tag is None else findtext(tag, "DisplayName")
         )
         self._storage_class = findtext(element, "StorageClass")
-        self._initiated_time = findtext(element, "Initiated")
-        if self._initiated_time:
-            self._initiated_time = from_iso8601utc(self._initiated_time)
+        initiated_time = findtext(element, "Initiated")
+        self._initiated_time = (
+            from_iso8601utc(initiated_time) if initiated_time else None
+        )
 
     @property
-    def object_name(self):
+    def object_name(self) -> str:
         """Get object name."""
         return self._object_name
 
     @property
-    def initiator_id(self):
+    def initiator_id(self) -> str | None:
         """Get initiator ID."""
         return self._initiator_id
 
     @property
-    def initator_name(self):
+    def initator_name(self) -> str | None:
         """Get initiator name."""
         return self._initiator_name
 
     @property
-    def owner_id(self):
+    def owner_id(self) -> str | None:
         """Get owner ID."""
         return self._owner_id
 
     @property
-    def owner_name(self):
+    def owner_name(self) -> str | None:
         """Get owner name."""
         return self._owner_name
 
     @property
-    def storage_class(self):
+    def storage_class(self) -> str | None:
         """Get storage class."""
         return self._storage_class
 
     @property
-    def upload_id(self):
+    def upload_id(self) -> str | None:
         """Get upload ID."""
         return self._upload_id
 
     @property
-    def initiated_time(self):
+    def initiated_time(self) -> datetime | None:
         """Get initiated time."""
         return self._initiated_time
 
@@ -554,7 +614,7 @@ class Upload:
 class ListMultipartUploadsResult:
     """ListMultipartUploads API result."""
 
-    def __init__(self, response):
+    def __init__(self, response: BaseHTTPResponse):
         element = ET.fromstring(response.data.decode())
         self._encoding_type = findtext(element, "EncodingType")
         self._bucket_name = findtext(element, "Bucket")
@@ -572,13 +632,11 @@ class ListMultipartUploadsResult:
                 if self._encoding_type == "url" else self._next_key_marker
             )
         self._next_upload_id_marker = findtext(element, "NextUploadIdMarker")
-        self._max_uploads = findtext(element, "MaxUploads")
-        if self._max_uploads:
-            self._max_uploads = int(self._max_uploads)
-        self._is_truncated = findtext(element, "IsTruncated")
+        max_uploads = findtext(element, "MaxUploads")
+        self._max_uploads = int(max_uploads) if max_uploads else None
+        is_truncated = findtext(element, "IsTruncated")
         self._is_truncated = (
-            self._is_truncated is not None and
-            self._is_truncated.lower() == "true"
+            is_truncated is not None and is_truncated.lower() == "true"
         )
         self._uploads = [
             Upload(tag, self._encoding_type)
@@ -586,47 +644,47 @@ class ListMultipartUploadsResult:
         ]
 
     @property
-    def bucket_name(self):
+    def bucket_name(self) -> str | None:
         """Get bucket name."""
         return self._bucket_name
 
     @property
-    def key_marker(self):
+    def key_marker(self) -> str | None:
         """Get key marker."""
         return self._key_marker
 
     @property
-    def upload_id_marker(self):
+    def upload_id_marker(self) -> str | None:
         """Get upload ID marker."""
         return self._upload_id_marker
 
     @property
-    def next_key_marker(self):
+    def next_key_marker(self) -> str | None:
         """Get next key marker."""
         return self._next_key_marker
 
     @property
-    def next_upload_id_marker(self):
+    def next_upload_id_marker(self) -> str | None:
         """Get next upload ID marker."""
         return self._next_upload_id_marker
 
     @property
-    def max_uploads(self):
+    def max_uploads(self) -> int | None:
         """Get max uploads."""
         return self._max_uploads
 
     @property
-    def is_truncated(self):
+    def is_truncated(self) -> bool:
         """Get is-truncated flag."""
         return self._is_truncated
 
     @property
-    def encoding_type(self):
+    def encoding_type(self) -> str | None:
         """Get encoding type."""
         return self._encoding_type
 
     @property
-    def uploads(self):
+    def uploads(self) -> list[Upload]:
         """Get uploads."""
         return self._uploads
 
@@ -644,7 +702,7 @@ _STARTS_WITH = "starts-with"
 _ALGORITHM = "AWS4-HMAC-SHA256"
 
 
-def _trim_dollar(value):
+def _trim_dollar(value: str) -> str:
     """Trim dollar character if present."""
     return value[1:] if value.startswith("$") else value
 
@@ -657,19 +715,19 @@ class PostPolicy:
     https://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-HTTPPOSTConstructPolicy.html#sigv4-PolicyConditions
     """
 
-    def __init__(self, bucket_name, expiration):
+    def __init__(self, bucket_name: str, expiration: datetime):
         check_bucket_name(bucket_name)
-        if not isinstance(expiration, datetime.datetime):
-            raise ValueError("expiration must be datetime.datetime type")
+        if not isinstance(expiration, datetime):
+            raise ValueError("expiration must be datetime type")
         self._bucket_name = bucket_name
         self._expiration = expiration
-        self._conditions = OrderedDict()
+        self._conditions: OrderedDict = OrderedDict()
         self._conditions[_EQ] = OrderedDict()
         self._conditions[_STARTS_WITH] = OrderedDict()
-        self._lower_limit = None
-        self._upper_limit = None
+        self._lower_limit: int | None = None
+        self._upper_limit: int | None = None
 
-    def add_equals_condition(self, element, value):
+    def add_equals_condition(self, element: str, value: str):
         """Add equals condition of an element and value."""
         if not element:
             raise ValueError("condition element cannot be empty")
@@ -686,13 +744,13 @@ class PostPolicy:
             raise ValueError(element + " cannot be set")
         self._conditions[_EQ][element] = value
 
-    def remove_equals_condition(self, element):
+    def remove_equals_condition(self, element: str):
         """Remove previously set equals condition of an element."""
         if not element:
             raise ValueError("condition element cannot be empty")
         self._conditions[_EQ].pop(element)
 
-    def add_starts_with_condition(self, element, value):
+    def add_starts_with_condition(self, element: str, value: str):
         """
         Add starts-with condition of an element and value. Value set to empty
         string does matching any content condition.
@@ -714,14 +772,14 @@ class PostPolicy:
             raise ValueError(element + " cannot be set")
         self._conditions[_STARTS_WITH][element] = value
 
-    def remove_starts_with_condition(self, element):
+    def remove_starts_with_condition(self, element: str):
         """Remove previously set starts-with condition of an element."""
         if not element:
             raise ValueError("condition element cannot be empty")
         self._conditions[_STARTS_WITH].pop(element)
 
     def add_content_length_range_condition(  # pylint: disable=invalid-name
-            self, lower_limit, upper_limit):
+            self, lower_limit: int, upper_limit: int):
         """Add content-length-range condition with lower and upper limits."""
         if lower_limit < 0:
             raise ValueError("lower limit cannot be negative number")
@@ -738,7 +796,7 @@ class PostPolicy:
         self._lower_limit = None
         self._upper_limit = None
 
-    def form_data(self, creds, region):
+    def form_data(self, creds: Credentials, region: str):
         """
         Return form-data of this post policy. The returned dict contains
         x-amz-algorithm, x-amz-credential, x-amz-security-token, x-amz-date,
@@ -754,7 +812,7 @@ class PostPolicy:
         ):
             raise ValueError("key condition must be set")
 
-        policy = OrderedDict()
+        policy: OrderedDict = OrderedDict()
         policy["expiration"] = to_iso8601utc(self._expiration)
         policy["conditions"] = [[_EQ, "$bucket", self._bucket_name]]
         for cond_key, conditions in self._conditions.items():
@@ -764,7 +822,7 @@ class PostPolicy:
             policy["conditions"].append(
                 ["content-length-range", self._lower_limit, self._upper_limit],
             )
-        utcnow = datetime.datetime.utcnow()
+        utcnow = datetime.utcnow()
         credential = get_credential_string(creds.access_key, utcnow, region)
         amz_date = to_amz_date(utcnow)
         policy["conditions"].append([_EQ, "$x-amz-algorithm", _ALGORITHM])
@@ -775,15 +833,17 @@ class PostPolicy:
             )
         policy["conditions"].append([_EQ, "$x-amz-date", amz_date])
 
-        policy = base64.b64encode(json.dumps(policy).encode()).decode("utf-8")
+        policy_encoded = base64.b64encode(
+            json.dumps(policy).encode(),
+        ).decode("utf-8")
         signature = post_presign_v4(
-            policy, creds.secret_key, utcnow, region,
+            policy_encoded, creds.secret_key, utcnow, region,
         )
         form_data = {
             "x-amz-algorithm": _ALGORITHM,
             "x-amz-credential": credential,
             "x-amz-date": amz_date,
-            "policy": policy,
+            "policy": policy_encoded,
             "x-amz-signature": signature,
         }
         if creds.session_token:
@@ -791,21 +851,19 @@ class PostPolicy:
         return form_data
 
     @property
-    def bucket_name(self):
+    def bucket_name(self) -> str:
         """Get bucket name."""
         return self._bucket_name
 
 
-def parse_copy_object(response):
+def parse_copy_object(
+        response: BaseHTTPResponse,
+) -> tuple[str, datetime | None]:
     """Parse CopyObject/UploadPartCopy response."""
     element = ET.fromstring(response.data.decode())
-    etag = findtext(element, "ETag")
-    if etag:
-        etag = etag.replace('"', "")
+    etag = cast(str, findtext(element, "ETag", True)).replace('"', "")
     last_modified = findtext(element, "LastModified")
-    if last_modified:
-        last_modified = from_iso8601utc(last_modified)
-    return etag, last_modified
+    return etag, from_iso8601utc(last_modified) if last_modified else None
 
 
 class EventIterable:
@@ -858,13 +916,19 @@ class EventIterable:
 class PeerSite:
     """Represents a cluster/site to be added to the set of replicated sites."""
 
-    def __init__(self, name, endpoint, access_key, secret_key):
+    def __init__(
+            self,
+            name: str,
+            endpoint: str,
+            access_key: str,
+            secret_key: str,
+    ):
         self._name = name
         self._endpoint = endpoint
         self._access_key = access_key
         self._secret_key = secret_key
 
-    def to_dict(self):
+    def to_dict(self) -> dict[str, str]:
         """Convert to dictionary."""
         return {
             "name": self._name,
@@ -897,86 +961,86 @@ class SiteReplicationStatusOptions:
         self._show_deleted = False
 
     @property
-    def buckets(self):
+    def buckets(self) -> bool:
         """Get buckets."""
         return self._buckets
 
     @buckets.setter
-    def buckets(self, value):
+    def buckets(self, value: bool):
         """Set buckets."""
         self._buckets = value
 
     @property
-    def policies(self):
+    def policies(self) -> bool:
         """Get policies."""
         return self._policies
 
     @policies.setter
-    def policies(self, value):
+    def policies(self, value: bool):
         """Set policies."""
         self._policies = value
 
     @property
-    def users(self):
+    def users(self) -> bool:
         """Get users."""
         return self._users
 
     @users.setter
-    def users(self, value):
+    def users(self, value: bool):
         """Set users."""
         self._users = value
 
     @property
-    def groups(self):
+    def groups(self) -> bool:
         """Get groups."""
         return self._groups
 
     @groups.setter
-    def groups(self, value):
+    def groups(self, value: bool):
         """Set groups."""
         self._groups = value
 
     @property
-    def metrics(self):
+    def metrics(self) -> bool:
         """Get metrics."""
         return self._metrics
 
     @metrics.setter
-    def metrics(self, value):
+    def metrics(self, value: bool):
         """Set metrics."""
         self._metrics = value
 
     @property
-    def entity(self):
+    def entity(self) -> str:
         """Get entity."""
         return self._entity
 
     @entity.setter
-    def entity(self, value):
+    def entity(self, value: str):
         """Set entity."""
         self._entity = value
 
     @property
-    def entity_value(self):
+    def entity_value(self) -> str:
         """Get entity value."""
         return self._entity_value
 
     @entity_value.setter
-    def entity_value(self, value):
+    def entity_value(self, value: str):
         """Set entity value."""
         self._entity_value = value
 
     @property
-    def show_deleted(self):
+    def show_deleted(self) -> bool:
         """Get show deleted."""
         return self._show_deleted
 
     @show_deleted.setter
-    def show_deleted(self, value):
+    def show_deleted(self, value: bool):
         """Set show deleted."""
         self._show_deleted = value
 
-    def to_query_params(self):
+    def to_query_params(self) -> dict[str, str]:
         """Convert this options to query parameters."""
         params = {
             "buckets": str(self._buckets).lower(),
@@ -995,83 +1059,88 @@ class SiteReplicationStatusOptions:
 class PeerInfo:
     """Site replication peer information."""
 
-    def __init__(self, deployment_id, endpoint, bucket_bandwidth_limit,
-                 bucket_bandwidth_set):
+    def __init__(
+            self,
+            deployment_id: str,
+            endpoint: str,
+            bucket_bandwidth_limit: str,
+            bucket_bandwidth_set: str,
+    ):
         self._deployment_id = deployment_id
         self._endpoint = endpoint
-        self._name = None
-        self._sync_status = None
+        self._name: str | None = None
+        self._sync_status: str | None = None
         self._bucket_bandwidth_limit = bucket_bandwidth_limit
         self._bucket_bandwidth_set = bucket_bandwidth_set
-        self._bucket_bandwidth_updated_at = None
+        self._bucket_bandwidth_updated_at: datetime | None = None
 
     @property
-    def deployment_id(self):
+    def deployment_id(self) -> str:
         """Get deployment ID."""
         return self._deployment_id
 
     @deployment_id.setter
-    def deployment_id(self, value):
+    def deployment_id(self, value: str):
         """Set deployment ID."""
         self._deployment_id = value
 
     @property
-    def endpoint(self):
+    def endpoint(self) -> str:
         """Get endpoint."""
         return self._endpoint
 
     @endpoint.setter
-    def endpoint(self, value):
+    def endpoint(self, value: str):
         """Set endpoint."""
         self._endpoint = value
 
     @property
-    def name(self):
+    def name(self) -> str | None:
         """Get name."""
         return self._name
 
     @name.setter
-    def name(self, value):
+    def name(self, value: str):
         """Set name."""
         self._name = value
 
     @property
-    def sync_status(self):
+    def sync_status(self) -> str | None:
         """Get sync status."""
         return self._sync_status
 
     @sync_status.setter
-    def sync_status(self, value):
+    def sync_status(self, value: str):
         """Set sync status."""
         self._sync_status = value
 
     @property
-    def bucket_bandwidth_limit(self):
+    def bucket_bandwidth_limit(self) -> str:
         """Get bucket bandwidth limit."""
         return self._bucket_bandwidth_limit
 
     @bucket_bandwidth_limit.setter
-    def bucket_bandwidth_limit(self, value):
+    def bucket_bandwidth_limit(self, value: str):
         """Set bucket bandwidth limit."""
         self._bucket_bandwidth_limit = value
 
     @property
-    def bucket_bandwidth_set(self):
+    def bucket_bandwidth_set(self) -> str:
         """Get bucket bandwidth set."""
         return self._bucket_bandwidth_set
 
     @bucket_bandwidth_set.setter
-    def bucket_bandwidth_set(self, value):
+    def bucket_bandwidth_set(self, value: str):
         """Set bucket bandwidth set."""
         self._bucket_bandwidth_set = value
 
     @property
-    def bucket_bandwidth_updated_at(self):
+    def bucket_bandwidth_updated_at(self) -> datetime | None:
         """Get bucket bandwidth updated at."""
         return self._bucket_bandwidth_updated_at
 
     @bucket_bandwidth_updated_at.setter
-    def bucket_bandwidth_updated_at(self, value):
+    def bucket_bandwidth_updated_at(self, value: datetime | None):
         """Set bucket bandwidth updated at."""
         self._bucket_bandwidth_updated_at = value
 
